@@ -2,13 +2,17 @@ use strict;
 use warnings;
 
 package Geo::Coordinates::UTM;
+
+use Carp;
+
 use base 'Exporter';
 
-our @EXPORT  = qw( latlon_to_utm  utm_to_latlon utm_to_mgrs
+our @EXPORT  = qw( latlon_to_utm  latlon_to_utm_force_zone
+                   utm_to_latlon utm_to_mgrs
                    latlon_to_mgrs mgrs_to_utm mgrs_to_latlon
 		   ellipsoid_info ellipsoid_names );
 
-our $VERSION = '0.06';
+our $VERSION = '0.11';
 
 use Math::Trig;
 my $deg2rad =  pi / 180;
@@ -36,11 +40,11 @@ BEGIN {  # Initialize this before other modules get a chance
     ( [ "Airy", 6377563, 0.00667054]
     , [ "Australian National", 6378160, 0.006694542]
     , [ "Bessel 1841", 6377397, 0.006674372]
-    , [ "Bessel 1841 (Nambia) ", 6377484, 0.006674372]
+    , [ "Bessel 1841 Nambia", 6377484, 0.006674372]
     , [ "Clarke 1866", 6378206, 0.006768658]
     , [ "Clarke 1880", 6378249, 0.006803511]
-    , [ "Everest", 6377276, 0.006637847]
-    , [ "Fischer 1960 (Mercury) ", 6378166, 0.006693422]
+    , [ "Everest 1830 India", 6377276, 0.006637847]
+    , [ "Fischer 1960 Mercury", 6378166, 0.006693422]
     , [ "Fischer 1968", 6378150, 0.006693422]
     , [ "GRS 1967", 6378160, 0.006694605]
     , [ "GRS 1980", 6378137, 0.00669438]
@@ -56,8 +60,26 @@ BEGIN {  # Initialize this before other modules get a chance
     , [ "WGS 66", 6378145, 0.006694542]
     , [ "WGS-72", 6378135, 0.006694318]
     , [ "WGS-84", 6378137, 0.00669438 ]
+    , [ "Everest 1830 Malaysia", 6377299, 0.006637847]
+    , [ "Everest 1956 India", 6377301, 0.006637847]
+    , [ "Everest 1964 Malaysia and Singapore", 6377304, 0.006637847]
+    , [ "Everest 1969 Malaysia", 6377296, 0.006637847]
+    , [ "Everest Pakistan", 6377296, 0.006637534]
+    , [ "Indonesian 1974", 6378160, 0.006694609]
+	, [ "Arc 1950", 6378249.145,0.006803481]
+	, [ "NAD 27",6378206.4,0.006768658]
+	, [ "NAD 83",6378137,0.006694384]
     );
 
+# calc ecc  as  
+# a = semi major axis
+# b = semi minor axis
+# e^2 = (a^2-b^2)/a^2	
+# For clarke 1880 (Arc1950) a=6378249.145 b=6356514.966398753
+# e^2 (40682062155693.23 - 40405282518051.34) / 40682062155693.23
+# e^2 = 0.0068034810178165
+	
+
   foreach my $el (@Ellipsoid)
   {   my ($name, $eqrad, $eccsq) = @$el;
       $Ellipsoid{$name} = $el;
@@ -101,21 +123,35 @@ sub ellipsoid_info($)
 sub latlon_to_utm($$$)
 {   my ($ellips, $latitude, $longitude) = @_;
 
-    my ($name, $radius, $eccentricity) = ellipsoid_info $ellips
-        or die "Ellipsoid value ($ellips) invalid.";
-
-    die "Longitude value ($longitude) invalid."
+    croak "Longitude value ($longitude) invalid."
         if $longitude < -180 || $longitude > 180;
 
-#>> original line contained +180 -180
-    my $long2       = $longitude - int(($longitude + 180)/360) * 360;
-    my $zone        = int( ($long2 + 180)/6) + 1;
+    my $long2 = $longitude - int(($longitude + 180)/360) * 360;
+    my $zone  = _latlon_zone_number($latitude, $long2);
 
-    my $lat_radian  = $deg2rad * $latitude;
-    my $long_radian = $deg2rad * $long2;
+    _latlon_to_utm($ellips, $zone, $latitude, $long2);
+}
 
-    my $k0          = 0.9996;   # scale
+sub latlon_to_utm_force_zone($$$$)
+{   my ($ellips, $zone, $latitude, $longitude) = @_;
 
+    croak "Longitude value ($longitude) invalid."
+        if $longitude < -180 || $longitude > 180;
+
+    my $long2 = $longitude - int(($longitude + 180)/360) * 360;
+
+    my ($zone_number) = $zone =~ /^(\d+)[CDEFGHJKLMNPQRSTUVWX]?$/i;
+
+    croak "Zone value ($zone) invalid."
+        unless defined($zone_number) && $zone_number <= 60;
+
+    _latlon_to_utm($ellips, $zone_number, $latitude, $long2);
+}
+
+sub _latlon_zone_number
+{   my  ($latitude, $long2) = @_;
+
+    my $zone = int( ($long2 + 180)/6) + 1;
     if($latitude >= 56.0 && $latitude < 64.0 && $long2 >= 3.0 && $long2 < 12.0)
     {   $zone = 32;
     }
@@ -126,6 +162,19 @@ sub latlon_to_utm($$$)
                  : ($long2 >= 33.0 && $long2 < 42.0) ? 37
 		 :                                     $zone;
     }
+    return $zone;
+}
+
+sub _latlon_to_utm
+{   my ($ellips, $zone, $latitude, $long2) = @_;
+
+    my ($name, $radius, $eccentricity) = ellipsoid_info $ellips
+        or croak "Ellipsoid value ($ellips) invalid.";
+
+    my $lat_radian  = $deg2rad * $latitude;
+    my $long_radian = $deg2rad * $long2;
+
+    my $k0          = 0.9996;   # scale
 
     my $longorigin       = ($zone - 1)*6 - 180 + 3;
     my $longoriginradian = $deg2rad * $longorigin;
@@ -178,7 +227,7 @@ sub latlon_to_utm($$$)
       : (-56 >  $latitude && $latitude >= -64) ? 'E'
       : (-64 >  $latitude && $latitude >= -72) ? 'D'
       : (-72 >  $latitude && $latitude >= -80) ? 'C'
-      : die "Latitude ($latitude) out of UTM range.";
+      : croak "Latitude ($latitude) out of UTM range.";
 
     $zone .= $utm_letter;
 
@@ -193,12 +242,12 @@ sub utm_to_latlon($$$$)
 {   my ($ellips, $zone, $easting, $northing) = @_;
 
     my ($name, $radius, $eccentricity) = ellipsoid_info $ellips
-        or die "Ellipsoid value ($ellips) invalid.";
+        or croak "Ellipsoid value ($ellips) invalid.";
        
     my $zone_number = $zone;
     my $zone_letter = chop $zone_number;
 
-    die "UTM zone ($zone_letter) invalid."
+    croak "UTM zone ($zone_letter) invalid."
        unless _valid_utm_zone $zone_letter;
 
     my $k0 = 0.9996;
@@ -238,11 +287,11 @@ sub utm_to_mgrs($$$)
    my $zone_number = $zone;
    my $zone_letter = chop $zone_number;
 
-   die "UTM zone ($zone_letter) invalid."
+   croak "UTM zone ($zone_letter) invalid."
      unless _valid_utm_zone $zone_letter;
 
    my $northing_zones="ABCDEFGHJKLMNPQRSTUV";
-   my $rnd_north=sprintf("%.0f",$northing);
+   my $rnd_north=sprintf("%d",$northing);
    my $north_split=length($rnd_north)-5;
    $north_split=0 if $north_split < 0;
    my $mgrs_north=substr($rnd_north,(length($rnd_north)-5));
@@ -253,7 +302,7 @@ sub utm_to_mgrs($$$)
    $num_north-=20 until $num_north <20;
    my $lett_north=substr($northing_zones,$num_north,1);
 
-   my $rnd_east=sprintf("%.0f",$easting);
+   my $rnd_east=sprintf("%d",$easting);
    my $east_split=length($rnd_east)-5;
    $east_split=0 if $east_split < 0;
    my $mgrs_east=substr($rnd_east,(length($rnd_east)-5));
@@ -267,9 +316,9 @@ sub utm_to_mgrs($$$)
       = ( $mgrs_zone == 1) ? 'ABCDEFGH'
       : ( $mgrs_zone == 2) ? 'JKLMNPQR'
       : ( $mgrs_zone == 3) ? 'STUVWXYZ'
-      : die "Could not calculate MGRS zone.";
+      : croak "Could not calculate MGRS zone.";
    $num_east--;
-   my $lett_east=substr($easting_zones,$num_east,1) or die "Could not detect Easting Zone for MGRS coordinate";
+   my $lett_east=substr($easting_zones,$num_east,1) or croak "Could not detect Easting Zone for MGRS coordinate";
 
    my $MGRS="$zone$lett_east$lett_north$mgrs_east$mgrs_north";
   ($MGRS);
@@ -285,46 +334,47 @@ sub latlon_to_mgrs($$$)
 
 sub mgrs_to_utm($)
 {  my ($mgrs_string) = @_;
-
-   my $zone = substr($mgrs_string,0,3);
+   my $zone = substr($mgrs_string,0,2);
+   $mgrs_string = "0$mgrs_string" if !($zone =~ /^\d+$/);
+   $zone = substr($mgrs_string,0,3);
    my $zone_number = $zone;
    my $zone_letter = chop $zone_number;
 
-   die "UTM zone ($zone_letter) invalid."
+   croak "UTM zone ($zone_letter) invalid."
      unless _valid_utm_zone $zone_letter;
 
    my $first_letter = substr($mgrs_string,3,1);
-   die "MGRS zone ($first_letter) invalid."
+   croak "MGRS zone ($first_letter) invalid."
      unless $first_letter =~ /[ABCDEFGHJKLMNPQRSTUVWXYZ]/;
 
    my $second_letter = substr($mgrs_string,4,1);
-   die "MGRS zone ($second_letter) invalid."
+   croak "MGRS zone ($second_letter) invalid."
      unless $second_letter =~ /[ABCDEFGHJKLMNPQRSTUV]/;
 
    my $coords=substr($mgrs_string,5);
    my $coord_len=length($coords);
-   die "MGRS coords ($coords) invalid."
+   croak "MGRS coords ($coords) invalid."
      unless ((($coord_len > 0) and ($coord_len <= 10)) and !($coord_len % 2));
    
    $coord_len=int($coord_len/2);
    my $x_coord=substr($coords,0,$coord_len);
    my $y_coord=substr($coords,$coord_len);
-   $x_coord*=10 until (length($x_coord) >= 5);
-   $y_coord*=10 until (length($y_coord) >= 5);
+   $x_coord*=10 ** (5 - $coord_len);
+   $y_coord*=10 ** (5 - $coord_len);
 
    my $east_pos
      = ( $first_letter =~ /[ABCDEFGH]/) ? index('ABCDEFGH',$first_letter)
      : ( $first_letter =~ /[JKLMNPQR]/) ? index('JKLMNPQR',$first_letter)
      : ( $first_letter =~ /[STUVWXYZ]/) ? index('STUVWXYZ',$first_letter)
-     : die "Could not calculate MGRS Easting zone.";
-   die "MGRS Letter $first_letter invalid." if $east_pos < 0;
+     : croak "Could not calculate MGRS Easting zone.";
+   croak "MGRS Letter $first_letter invalid." if $east_pos < 0;
    $east_pos++;
    $east_pos*=100000;
    $x_coord+=$east_pos;
 
    my $northing_zones="ABCDEFGHJKLMNPQRSTUV";
    my $north_pos=index($northing_zones,$second_letter);
-   die "MGRS Letter $second_letter invalid." if $north_pos < 0;
+   croak "MGRS Letter $second_letter invalid." if $north_pos < 0;
    $north_pos++;
    $north_pos-=5 if not ($zone_number % 2);
    $north_pos+=20 until $north_pos > 0;
@@ -334,7 +384,7 @@ sub mgrs_to_utm($)
        $tmpNorth++;
        $tmpNorth*=8;
        $tmpNorth*=10/9;
-       $tmpNorth=int(($tmpNorth-$north_pos)/20)*20;
+       $tmpNorth=int((($tmpNorth-$north_pos)/20)+0.5)*20;
        $north_pos+=$tmpNorth;
        $north_pos*=100000;
        $north_pos-=100000;
@@ -342,16 +392,14 @@ sub mgrs_to_utm($)
    }
    else {
        #Southern Hemisphere
-       #my $tmpNorth=index('MLKJHGFEDC',$zone_letter);
        my $tmpNorth=index('CDEFGHJKLM',$zone_letter);
-       $tmpNorth++;
        $tmpNorth*=8;
        $tmpNorth*=10/9;
-       $tmpNorth=int(($tmpNorth-$north_pos)/20)*20;
+       $tmpNorth=int((($tmpNorth-$north_pos)/20)+0.5)*20;
        $north_pos+=$tmpNorth;
        $north_pos*=100000;
        $north_pos-=100000;
-       $north_pos+=2000000;
+       $north_pos+=2000000 if $zone_letter ne "C";
        $y_coord+=$north_pos;
    }
 
@@ -448,7 +496,7 @@ The Ellipsoids available are as follows:
 
 =item 6 Clarke 1880
 
-=item 7 Everest
+=item 7 Everest 1830 (India)
 
 =item 8 Fischer 1960 (Mercury)
 
@@ -482,6 +530,24 @@ The Ellipsoids available are as follows:
 
 =item 23 WGS-84
 
+=item 24 Everest 1830 (Malaysia)
+
+=item 25 Everest 1956 (India)
+
+=item 26 Everest 1964 (Malaysia and Singapore)
+
+=item 27 Everest 1969 (Malaysia)
+
+=item 28 Everest (Pakistan)
+
+=item 29 Indonesian 1974
+
+=item 30 Arc 1950
+
+=item 30 NAD 27
+
+=item 30 NAD 83
+
 =back
 
 
@@ -520,6 +586,40 @@ returns
      $east  = 512543.777159849
      $north = 6406592.20049111
 
+=head2 latlon_to_utm_force_zone
+
+On occasions, it is necessary to map a pair of (latitude, longitude)
+coordinates to a predefined zone. This function allows to select the
+projection zone as follows:
+
+     ($zone, $east, $north)=latlon_to_utm('international', $zone_number,
+                                          $latitude, $longitude)
+
+For instance, Spain territory goes over zones 29, 30 and 31 but
+sometimes it is convenient to use the projection corresponding to zone
+30 for all the country.
+
+Santiago de Compostela is at 42deg 52min 57.06sec North, 8deg 32min 28.70sec West
+
+    ($zone, $east, $norh)=latlon_to_utm('international',  42.882517, -8.541306)
+
+returns
+
+     $zone = 29T
+     $east = 537460.331
+     $north = 4747955.991
+
+but forcing the conversion to zone 30:
+
+    ($zone, $east, $norh)=latlon_to_utm_force_zone('international',
+                                                   30, 42.882517, -8.541306)
+
+returns
+
+    $zone = 30T
+    $east = 47404.442
+    $north = 4762771.704
+
 =head2 utm_to_latlon
 
 Reversing the above example,
@@ -565,6 +665,26 @@ returns
      $latitude  = 57.8030590197684
      $longitude = -2.788956799645
 
+=head2 mgrs_to_utm
+
+    Similarly it is possible to convert MGRS directly to UTM
+
+        ($zone,$easting,$northing)=mgrs_to_utm('30VWK1254306804')
+
+    returns
+
+        $zone = 30V
+        $easting = 512543
+        $northing = 6406804
+
+=head2 utm_to_mgrs
+
+    and the inverse converting from UTM yo MGRS is done as follows
+
+       ($mgrs)=utm_to_mgrs('30V',512543,6406804);
+
+    returns
+        $mgrs = 30VWK1254306804
 
 =head1 AUTHOR
 
@@ -582,9 +702,17 @@ Mark Overmeer for the ellipsoid_info routines and code review.
 
 Lok Yan for the >72deg. N bug.
 
+Salvador Fandino for the forced zone UTM and additional tests
+
+Matthias Lendholt for modifications to MGRS calculations
+
+Peder Stray for the short MGRS patch
+
+
+
 =head1 COPYRIGHT
 
-Copyright (c) 2000,2002,2004,2007 by Graham Crookham.  All rights reserved.
+Copyright (c) 2000,2002,2004,2007,2010,2013 by Graham Crookham.  All rights reserved.
     
 This package is free software; you can redistribute it
 and/or modify it under the same terms as Perl itself.             
